#Matthew Aronson
#CHE210D Final Project
#June 15, 2012

#import standard modules
import numpy as np
import glob
import os
import atomwrite

#3D visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis

#import compiled fortran library
import finalproj

def ReadXtl(XtlFile):
    """reads .xtl file and outputs zeolite lattice information
Input:
    XtlFile: fractional coordinates of zeolite lattice
Output:
    Names: element name of each atom in zeolite lattice
    Pos: absolute (Angstrom) positions of atoms in zeolite lattice
    L: zeolite unit cell parameters
"""
    #create empty list and array to be filled and returned by the function
    Names = []
    Pos = np.array([],float)
    L = np.zeros((1,3),float)
    
    #line counters for reading the .xtl file
    linecount = 1
    abcline = 3
    xyzline = 8
    
    #read through all of the lines in the selected .xtl file
    for line in file(XtlFile,"r"):
        if linecount == abcline:
            L[0][0] = float(line[1:10])
            L[0][1] = float(line[12:21])
            L[0][2] = float(line[23:32])
        if linecount >= xyzline:
            if "EOF" in line:
                break
            elif "O" in line:
                Names.append(line[0:1])
                Pos = np.append(Pos,[float(line[13:21]), float(line[24:32]), float(line[35:42])])
            elif "Si" in line:
                Names.append(line[0:2])
                Pos = np.append(Pos,[float(line[13:21]), float(line[24:32]), float(line[35:42])])
        linecount += 1
    
    Pos = np.reshape(Pos,(len(Pos)/3.0,3))
    
    return Names, Pos, L


def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #identify global variables
    global LatA, NLat, M, L, Cut
    
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = np.clip(Dir, -1.e100, 1.e100)
    NormDir = NormDir / np.sqrt(np.sum(NormDir * NormDir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [finalproj.calcenergy(Pos + NormDir * x, LatA, M, L, Cut, NLat) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = finalproj.calcenergy(Pos + NormDir * dx, LatA, M, L, Cut, NLat)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [finalproj.calcenergy(Pos + NormDir * 2. * dx, LatA, M, L, Cut, NLat)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [finalproj.calcenergy(Pos + NormDir * Dists[-1], LatA, M, L, Cut, NLat)]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = finalproj.calcenergy(Pos + NormDir * d3, LatA, M, L, Cut, NLat)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = finalproj.calcenergy(Pos + NormDir * d3, LatA, M, L, Cut, NLat)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]
    
    #append Lattice positions w/ PosMin
    LatPos = np.append(LatA, PosMin, axis=0)
    
    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(LatPos, L)
        else:
            #initialize the visualization window
            atomvis.Init(LatPos, L)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    #identify global variables
    global Lat, NLat, M, L, Cut, Dim, N
    
    #initial search direction
    Forces = np.zeros_like(Pos)
    PE, Forces = finalproj.calcenergyforces(Pos, LatA, M, L, Cut, Forces, NLat)
    #PE, Forces = finalproj.calcenergyforces(Pos, Lat, NLat, M, L, Cut, Forces)
    Dir = Forces
    OldPE = 1.e300
    #iterative line searches
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = finalproj.calcenergyforces(Poly, Lat, M, L, Cut, Forces, NLat)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos  


def Berendsen(Vel,T,Tinst,dt):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
    Tinst: instantaneous temperature
    dt: time step
Output:
    Vel: same as above
"""
    Tau=dt/0.01
    Vel = Vel - Vel.mean(axis=0)
    Lambda = np.sqrt(1+(dt/Tau)*(T/Tinst - 1))
    Vel = Vel * Lambda
    return Vel


def InitVelocities(N, T, dt):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
    dt: time step
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = Berendsen(Vel, T, InstTemp(Vel),dt)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #global variables
    global Lat, NLat, M, L, Cut
    #get the acceleration from the forces
    Forces = np.zeros_like(Pos)
    PEnergy, Accel = finalproj.calcenergyforces(Pos, Lat, M, L, Cut, Forces, NLat)
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))
    
    
def Histogram(CoM, r, b):
    """Returns a histogram of the shortest distance from polymer atom to lattice
Input:
    CoM: polymer chain center-of-mass
    r: zeolite supercage radius
    b: number of bins
Output:
    prox: closest distance from polymer atom to lattice atom
"""
    global LatA
    
    #construct histogram bins
    dz=r/float(b) #bin width
    bin=[(0+i*dz) for i in range(0, b+1)]
    
    #initialize shortest distance value and array to hold histogram counts
    ci = 1000.
    prox = np.zeros((len(bin)-1), int)

    #find shortest distance from polymer CoM to lattice
    for latpos in LatA:
        dist = np.sqrt(np.sum((CoM - latpos)**2))
        if (dist < ci):
            ci = dist
    #initialize counter to loop through bin sets
    k = 0
    #iterate through all the bins
    while k<(len(bin)-1):
        if ci >= bin[k] and ci < bin[k+1]: #check to see if the distance is within a particular bin
            prox[k] += 1 #add a count in bin k
        k += 1 #increment the bin index

    return bin, prox


#========== GLOBAL VARIABLES ==========#
Cut = 4.0 #distance cutoff for pairwise interactions
Tlist = [1.0, 1.5, 2.0, 2.5, 3.0] #list of target temperatures
Mlist = [4, 6, 8] #list of polymer chain lengths
RescaleFreq = 100
dy = -1.0 #increment for initial polymer atom spacings

#========== MD parameters ==========#  
NStepsEquil1 = 10000 #number of steps for first equilibration period
NStepsEquil2 = 10000 #number of steps for second equilibration period
dt = 0.001 #time step
ttot = 100 #total time for NStepsProd = 100/0.001 = 100000
NStepsProd = 100000

#initialize variables for building zeolite lattice
LatA = np.array([],float)
L = np.array([12.632, 12.632, 26.186])
NLat = 0.

#specify zeolite framework to simulate
zeolite = "BEA"

#read .xtl file of zeoltie specified above & and extract lattice information
for XtlLat in glob.glob(os.path.join(".",zeolite + ".xtl")):
    Names, Lat, L = ReadXtl(XtlLat)
    
    #number of atoms in lattice
    NLat = len(Lat)
    #angstrom coordinates of BEA zeolite lattice
    LatA = Lat*L
    #write lattice positions to pdb file for Chimera
    atomwrite.pdbfile(zeolite + ".pdb", L, AtomNames = Names, Compressed = False).write(LatA)

#MD simulation loops
for M in Mlist:
    N = M #place only one polymer of length M in the lattice such that N = M
    
    #initialize lists for placing polymer atoms in zeolite lattice
    Poly = np.zeros((M,3),float)
    PolyXYZ = np.zeros((M,3),float)
    PolyNames = []
    #place first polymer atom in zeolite void & assign name as carbon atom
    Poly[0,:] = [11.000, 14.053, 6.537] #2
    PolyNames.append('C')
    #increment additional atom positions by amount dy
    for i in range (1,M):
        Poly[i,:] = Poly[0,:] + [0., i*dy, 0.]
        PolyNames.append('C')
    #append position and names arrays
    InitPos = np.append(LatA, Poly, axis=0)
    Names.extend(PolyNames)
    
    #loop thru all temperature values
    for T in Tlist:
        #construct file name from zeolite type, chain length, & temperature
        filename = zeolite + "anim_" + str(M) + "_" + str(T) + ".pdb"
        #open pdb file & set frequency to write frames
        Pdb = atomwrite.pdbfile(filename, L)
        PdbFreq = 100
        #energy-minimize the initial configuration using conjugate gradient method
        dx = 0.001
        EFracTolLS = 1.*10**(-8)
        EFracTolCG = 1.*10**(-10)
        PEnergy, Pos = ConjugateGradient(Poly, dx, EFracTolLS, EFracTolCG)
        #initial parameters for velocity Verlet integrator
        Accel = InitAccel(Pos)
        Vel = InitVelocities(N, T, dt)
        InTemp = InstTemp(Vel)
        
        #initialize loop counter
        niter = 0
        
        #NStepsEquil1 equilibration loop
        while niter <= NStepsEquil1:
            Pos, Vel, Accel, KEnergy, PEnergy = finalproj.vvintegrate(Pos, LatA, Vel, Accel, M, L, Cut, dt, NLat)  
            #check for velocity rescaling step number
            if niter % RescaleFreq == 0:
                Vel = Berendsen(Vel, T, InstTemp(Vel), dt)
            #increment loop counter
            niter += 1
        
        #create array for energy sum (kinetic + potential)
        TotE = np.array([],float)
        #reset loop counter
        niter = 0
        
        #NStepsEquil2 equilibration loop
        while niter <= NStepsEquil2:
            Pos, Vel, Accel, KEnergy, PEnergy = finalproj.vvintegrate(Pos, LatA, Vel, Accel, M, L, Cut, dt, NLat)  
            #check for velocity rescaling step number
            if niter % RescaleFreq == 0:
                Vel = Berendsen(Vel, T, InstTemp(Vel), dt)
            #update energy array
            TotE = np.append(TotE, KEnergy + PEnergy)
            #increment loop counter
            niter += 1
        
        #compute average energy of NStepsEquil2 equilibration period
        AvgE = np.mean(TotE)
        #find the target kinetic energy
        KEtar = AvgE - PEnergy
        #velocity rescale factor
        Vrscl = np.sqrt(KEtar / KEnergy)
        #rescale velocity
        Vel = Vel * Vrscl
        #copy particle positions for calculating the mean-squared displacement
        Pos0 = Pos.copy()
        
        #initialize time counter and reset loop counter
        t = 0.
        niter = 0
        
        #construct histogram
        nbins = 20
        radius = 7.
        Hist = np.zeros((nbins), int) #histogam list
        HistCount = 0.
        
        #print header for output data
        print "M = " + str(M)
        print "T = " + str(T)
        print "%4s %7s" %("Time", "MSD")
        print "-"*18
        
        #Production run
        while niter <= NStepsProd:
            Pos, Vel, Accel, KEnergy, PEnergy = finalproj.vvintegrate(Pos, LatA, Vel, Accel, M, L, Cut, dt, NLat)  
            #periodically record time and mean-squared displacement
            if niter % 1000 == 0:
                MSD = np.sum((Pos-Pos0)*(Pos-Pos0))/len(Pos)
                print "%5.1f %10.5f" %(t, MSD)
                #calculate polymer chain center-of-mass (CoM)
                CoM = np.mean(Pos, axis=0)      
                #add histogram counts to bins
                bin, prox = Histogram(CoM, radius, nbins)
                Hist += prox
                HistCount += 1
            #append lattice positions w/ PosMin
            LatPos = np.append(LatA, Pos, axis=0)
            #3D visualization
            if UseVisual:
                if atomvis.Initialized:
                    #update the positions
                    atomvis.Update(LatPos, L)
                else:
                    #initialize the visualization window
                    atomvis.Init(LatPos, L) 
            #save frame as pdb file
            if niter % PdbFreq == 0:
                Pdb.write(Pos)
                    
            #increment counters
            t += dt
            niter += 1
        
        print Hist/HistCount #normalized bin counts (probabilities)
        print bin #bin widths
        print np.sum(Hist/HistCount) #check to make sure = 1
        print "-"*36 
        print "-"*36 #separate MD runs
    
        #close pdb file
        Pdb.close()